"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.GetDefinedTypeHistogramVisitor = void 0;
const Visitor_1 = require("../Visitor");
class GetDefinedTypeHistogramVisitor {
    constructor() {
        this.mode = null;
        this.stackLevel = 0;
    }
    visitRoot(root) {
        return this.mergeHistograms(root.programs.map((program) => (0, Visitor_1.visit)(program, this)));
    }
    visitProgram(program) {
        return this.mergeHistograms([
            ...program.accounts.map((account) => (0, Visitor_1.visit)(account, this)),
            ...program.instructions.map((instruction) => (0, Visitor_1.visit)(instruction, this)),
            ...program.definedTypes.map((type) => (0, Visitor_1.visit)(type, this)),
        ]);
    }
    visitAccount(account) {
        this.mode = 'account';
        this.stackLevel = 0;
        const histogram = (0, Visitor_1.visit)(account.data, this);
        this.mode = null;
        return histogram;
    }
    visitAccountData(accountData) {
        return (0, Visitor_1.visit)(accountData.struct, this);
    }
    visitInstruction(instruction) {
        this.mode = 'instruction';
        this.stackLevel = 0;
        const dataHistogram = (0, Visitor_1.visit)(instruction.dataArgs, this);
        const extraHistogram = (0, Visitor_1.visit)(instruction.extraArgs, this);
        this.mode = null;
        const subHistograms = instruction.subInstructions.map((ix) => (0, Visitor_1.visit)(ix, this));
        return this.mergeHistograms([
            dataHistogram,
            extraHistogram,
            ...subHistograms,
        ]);
    }
    visitInstructionAccount() {
        return {};
    }
    visitInstructionDataArgs(instructionDataArgs) {
        return (0, Visitor_1.visit)(instructionDataArgs.struct, this);
    }
    visitInstructionExtraArgs(instructionExtraArgs) {
        return (0, Visitor_1.visit)(instructionExtraArgs.struct, this);
    }
    visitDefinedType(definedType) {
        this.mode = 'definedType';
        this.stackLevel = 0;
        const histogram = (0, Visitor_1.visit)(definedType.data, this);
        this.mode = null;
        return histogram;
    }
    visitError() {
        return {};
    }
    visitArrayType(arrayType) {
        this.stackLevel += 1;
        const histogram = (0, Visitor_1.visit)(arrayType.child, this);
        this.stackLevel -= 1;
        return histogram;
    }
    visitLinkType(linkType) {
        if (linkType.importFrom !== 'generated') {
            return {};
        }
        return {
            [linkType.name]: {
                total: 1,
                inAccounts: Number(this.mode === 'account'),
                inDefinedTypes: Number(this.mode === 'definedType'),
                inInstructionArgs: Number(this.mode === 'instruction'),
                directlyAsInstructionArgs: Number(this.mode === 'instruction' && this.stackLevel <= 2),
            },
        };
    }
    visitEnumType(enumType) {
        this.stackLevel += 1;
        const histogram = this.mergeHistograms(enumType.variants.map((variant) => (0, Visitor_1.visit)(variant, this)));
        this.stackLevel -= 1;
        return histogram;
    }
    visitEnumEmptyVariantType() {
        return {};
    }
    visitEnumStructVariantType(enumStructVariantType) {
        this.stackLevel += 1;
        const histogram = (0, Visitor_1.visit)(enumStructVariantType.struct, this);
        this.stackLevel -= 1;
        return histogram;
    }
    visitEnumTupleVariantType(enumTupleVariantType) {
        this.stackLevel += 1;
        const histogram = (0, Visitor_1.visit)(enumTupleVariantType.tuple, this);
        this.stackLevel -= 1;
        return histogram;
    }
    visitMapType(mapType) {
        this.stackLevel += 1;
        const histogram = this.mergeHistograms([
            (0, Visitor_1.visit)(mapType.key, this),
            (0, Visitor_1.visit)(mapType.value, this),
        ]);
        this.stackLevel -= 1;
        return histogram;
    }
    visitOptionType(optionType) {
        this.stackLevel += 1;
        const histogram = (0, Visitor_1.visit)(optionType.child, this);
        this.stackLevel -= 1;
        return histogram;
    }
    visitSetType(setType) {
        this.stackLevel += 1;
        const histogram = (0, Visitor_1.visit)(setType.child, this);
        this.stackLevel -= 1;
        return histogram;
    }
    visitStructType(structType) {
        this.stackLevel += 1;
        const histogram = this.mergeHistograms(structType.fields.map((field) => (0, Visitor_1.visit)(field, this)));
        this.stackLevel -= 1;
        return histogram;
    }
    visitStructFieldType(structFieldType) {
        this.stackLevel += 1;
        const histogram = (0, Visitor_1.visit)(structFieldType.child, this);
        this.stackLevel -= 1;
        return histogram;
    }
    visitTupleType(tupleType) {
        this.stackLevel += 1;
        const histogram = this.mergeHistograms(tupleType.children.map((child) => (0, Visitor_1.visit)(child, this)));
        this.stackLevel -= 1;
        return histogram;
    }
    visitBoolType() {
        return {};
    }
    visitBytesType() {
        return {};
    }
    visitNumberType() {
        return {};
    }
    visitNumberWrapperType(numberWrapperType) {
        this.stackLevel += 1;
        const histogram = (0, Visitor_1.visit)(numberWrapperType.number, this);
        this.stackLevel -= 1;
        return histogram;
    }
    visitPublicKeyType() {
        return {};
    }
    visitStringType() {
        return {};
    }
    mergeHistograms(histograms) {
        const result = {};
        histograms.forEach((histogram) => {
            Object.keys(histogram).forEach((key) => {
                if (result[key] === undefined) {
                    result[key] = histogram[key];
                }
                else {
                    result[key].total += histogram[key].total;
                    result[key].inAccounts += histogram[key].inAccounts;
                    result[key].inDefinedTypes += histogram[key].inDefinedTypes;
                    result[key].inInstructionArgs += histogram[key].inInstructionArgs;
                    result[key].directlyAsInstructionArgs +=
                        histogram[key].directlyAsInstructionArgs;
                }
            });
        });
        return result;
    }
}
exports.GetDefinedTypeHistogramVisitor = GetDefinedTypeHistogramVisitor;
//# sourceMappingURL=GetDefinedTypeHistogramVisitor.js.map